#ifndef PREPROCESSOR_GRAPH_H_
#define PREPROCESSOR_GRAPH_H_

#include <vector>
#include <algorithm>

#include "lib_sharpsat_td/bitset.hpp"
#include "core/SolverTypes.h"

class Graph {
public:
	Graph() : nodes(0), edges(0) { }
	Graph(int vars, const std::vector<std::vector<Glucose::Lit>>& clauses);
	Graph(int vars, const std::vector<std::vector<Glucose::Lit>>& clauses, const std::vector<std::vector<Glucose::Lit>>& learnts, std::vector<int>& freq);

	void init(int n);
	void clear();

	int numNodes() { return this->nodes; }
	int numEdges() { return this->edges; }

	void addEdge(int v1, int v2);
	bool hasEdge(int v1, int v2) { return adj_mat[v1].Get(v2); }
	const std::vector<int> Neighbors(int v) const { return adj_list[v]; }

	bool isSimplical(int v) { return isClique(adj_list[v]); }
	bool isClique(const std::vector<int>& adj);

	// Debug
	void toDimacs(std::ostream& out, bool withHeader=true);

protected:
	int nodes;
	int edges;
	std::vector<std::vector<int>> adj_list;
	std::vector<sspp::Bitset> adj_mat;
};

#endif /* PREPROCESSOR_GRAPH_H_ */
