#include "counter/Config.h"

using namespace GraphPC;

//=================================================================================================
// Options:

static const char* omain = "GraphPC -- MAIN";
static IntOption		opt_vs			(omain, "vs", "Variable Selection Heuristics", 1, IntRange(0,1));
static BoolOption		opt_cache		(omain, "cc", "Use cache", true);
static IntOption		opt_cachesize	(omain, "cs", "Maximum component cache size (MB) (not strict)", 4000, IntRange(1, INT32_MAX));
static BoolOption		opt_bj			(omain, "bj", "Backjumping", false);
static DoubleOption	opt_bj_thd		(omain, "bjthd", "Backjumping threshold", 0.5, DoubleRange(0, false, 1, false));
static BoolOption		opt_simp		(omain, "rmvsatcl", "Remove satisfied clauses", true);
static IntOption		opt_simp_thd	(omain, "rmvsatclthd", "Threshold of removing satisfied clauses", 2, IntRange(0,INT32_MAX));
static BoolOption		opt_restart	(omain, "restart", "Restarting", false);
static StringOption	opt_mapout		(omain, "mapout", "Outfile for map between the original input and the preprocessed one.", "NULL");
static StringOption	opt_graphout	(omain, "gout", "Outfile for graph.", "NULL");
static BoolOption		opt_allpairs	(omain, "allpairs", "Count paths of all pairs", false);
static BoolOption		opt_progress	(omain, "progress", "Progress bar for allpairs counting", false);
// static BoolOption		opt_pp			(omain, "pp", "Preprocessing", true);

static const char* opp = "GraphPC -- Preprocessor";
static StringOption	opt_ppout		(opp, "ppout", "Outfile for Simplified CNF", "NULL");
static IntOption		opt_varlimit	(opp, "varlim", "limit on #Vars in Preprocessing.", 200000, IntRange(0, INT32_MAX));
static DoubleOption	opt_pptimelim	(opp, "pptimelim", "Time threshold of preprocessing (not precise).", 120, DoubleRange(0, true, DBL_MAX, true));
static IntOption		opt_pprep		(opp, "ppreps",	"#reps of the main loop of preprocessing.", 20, IntRange(1, INT32_MAX));
static IntOption		opt_ppverb		(opp, "ppverb", "Preprocessing verbosity level (0=some, 1=more).", 0, IntRange(0, 1));
static BoolOption		opt_ee			(opp, "pp_ee", "Use equivalent literal elimination.", true);
static IntOption		opt_ee_varlim	(opp, "ee-varlim", "limit on #Vars in equivalent literal elimination.", 150000, IntRange(0, INT32_MAX));
static BoolOption		opt_ve			(opp, "pp_ve", "Use variable elimination.", true);
static IntOption		opt_vereps		(opp, "ve_reps", "VE: the number of repetitions", 400, IntRange(0, INT32_MAX));
static IntOption		opt_dvereps	(opp, "dve_reps", "DefVE: the number of repetitions", 10, IntRange(0, INT32_MAX));
static BoolOption		opt_vemore		(opp, "ve_more", "VE: target more variables", true);
static DoubleOption	opt_dvetimelim(opp, "dve_timelim", "DefVE: time limit", 60.0, DoubleRange(0, true, DBL_MAX, true));
static BoolOption		opt_cs			(opp, "pp_cs", "Clause Strengthening", true);
static IntOption		opt_sat_bdj	(opp, "pp_sat_bdj", "conflict budget for the last sat solving", 50000, IntRange(1, INT32_MAX));

//=================================================================================================
Configuration::Configuration() {
	cntr.backjump = opt_bj;
	cntr.bj_threshold = opt_bj_thd;
	cntr.remove_sat_cls = opt_simp;
	cntr.rmvsatcl_threshold = opt_simp_thd;
	cntr.restart = opt_restart;
	cntr.cache = opt_cache;
	cntr.cachesize = opt_cachesize;
	cntr.pp_outfile = opt_ppout;
	cntr.map_outfile = opt_mapout;
	cntr.graph_outfile = opt_graphout;
	cntr.allpairs = opt_allpairs;
	cntr.progressbar = opt_progress;
//	cntr.preprocess = opt_pp;

	pp.varlimit = opt_varlimit;
	pp.timelim = opt_pptimelim;
	pp.reps = opt_pprep;
	pp.verb = opt_ppverb;
	pp.ee = opt_ee;
	pp.ee_varlim = opt_ee_varlim;
	pp.ve = opt_ve;
	pp.ve_reps = opt_vereps;
	pp.dve_reps = opt_dvereps;
	pp.ve_more = opt_vemore;
	pp.dve_timelimit = opt_dvetimelim;
	pp.cs = opt_cs;
	pp.confl_budget = opt_sat_bdj;
}
